############################################################################
##
## Copyright (c) 2000-2001 BalaBit IT Ltd, Budapest, Hungary
## All rights reserved.
##
## $Id: Chainer.py,v 1.38.2.4.2.1.2.5 2002/04/26 12:14:25 bazsi Exp $
##
############################################################################

"""Module defining the chainer interface for chaining up to parent proxies.

This module defines the 'AbstractChainer' interface, whose implementations
are responsible to connect a child proxy to its parent.
"""

from Zorp import *
from Connector import Connect
from Stream import Stream
from Session import StackedSession
from NAT import NAT_SNAT, NAT_DNAT
import types

class AbstractChainer:
	"""Abstract class encapsulating the chaining process.
	
	This abstract class encapsulates the process of chaining to
	the parent proxy.
	"""
	def __init__(self):
		pass

	def chainParent(self, session):
		"""Function to be called when a proxy wants to connect to its parent.
		
		This function is called to actually perform chaining to the parent.
		
		Arguments
		
		  self -- this instance
		  
		  session -- session we belong to
		"""
		raise NotImplementedError

class ConnectChainer(AbstractChainer):
	"""Class encapsulating a connection establishment.
	
	This class encapsulates a real TCP/IP connection establishment, and
	is used when a top-level proxy wants to perform chaining.
	
	When no explicit top-level chainer is specified for services, this
	one is used by default.	
	"""

	def establishConnection(self, session, local, remote):
		"""Function to actually establish a connection.
		
		Internal function to establish a connection with the given
		local and remote addresses. It is used by derived chainer
		classes after finding out the destination address to connect
		to. This function performs access control checks.
		   
		Arguments

		  self         -- this instance
		  
		  session      -- Session object
		  
		  local	       -- bind address
		  
		  remote       -- host to connect to
		  
		Returns
		
		  The stream of the connection to the server

		"""
		try:
			session.setServer(remote)
			if session.isServerPermitted() == Z_ACCEPT:
				#remote.options = session.client_address.options
				conn = Connect(local, remote)
				session.server_fd = conn.blockingConnect()
				session.server_local = conn.local
				if session.server_fd == -1:
					session.server_stream = None
					log(CORE_SESSION, 3, "%s: Server connection failure; server_addr='%s', server_zone='%s', server_local='%s'" % (session.session_id, session.server_address, session.server_zone, session.server_local))
				else:
					session.server_stream = Stream(session.server_fd, session.owner.session_id + ": server")
					log(CORE_SESSION, 3, "%s: Server connection established; server_fd='%d', server_addr='%s', server_zone='%s', server_local='%s'" % (session.session_id, session.server_fd, session.server_address, session.server_zone, session.server_local))
				return session.server_stream
			raise DACException
		except DACException:
			pass
		except MACException:
			pass
		return None

	def chainParent(self, session):
		"""Overridden function to perform connection establishment.
		
		This function is called by the actual Proxy implementation
		to actually connect to the server-endpoint of the session. 
		The destination address is 'session.server_address' (which
		is previously set by the Router used for the service, and
		optionally overridden by the Proxy). The local address to
		bind to is determined with the help of a NAT object if one
		is provided, or allocated dynamically by the kernel.
		
		Arguments
		
		  self -- this instance
		  
		  session -- session we belong to
		
		"""
		local = None
		remote = session.server_address
		if remote == None:
			log(CORE_SESSION, 3, "%s: Server connection failure, no destination;" % (session.session_id))
			return None

		if session.service.snat:
			local = session.service.snat.performTranslation(session, local, NAT_SNAT)

		if session.service.dnat:
			remote = session.service.dnat.performTranslation(session, remote, NAT_DNAT)
		return self.establishConnection(session, local, remote)

class FailoverChainer(ConnectChainer):
	"""Class encapsulating a connection establishment with failover capability.
	
	This class encapsulates a real TCP/IP connection establishment, and
	is used when a top-level proxy wants to perform chaining. In 
        addition to ConnectChainer this class adds the capability to perform
	failover, given a destination is not working.

	Attributes

	  current_host -- index of the last attempted destination
	"""
	def __init__(self):
		"""Constructor to initialize a FailoverChainer class.

		This constructor initializes a FailoverChainer class by
		filling attributes with appropriate values and calling the
		inherited constructor.

		Arguments

		  self -- this instance
		"""
		ConnectChainer.__init__(self)
		self.current_host = 0

	def chainParent(self, session):
		"""Overridden function to perform connection establishment.
		
		This function is called by the actual Proxy implementation
		to actually connect to the server-endpoint of the session. 
		The destination address is 'session.server_address' (which
		is previously set by the Router used for the service, and
		optionally overridden by the Proxy). The local address to
		bind to is determined with the help of a NAT object if one
		is provided, or allocated dynamically by the kernel.

		The failover capability of FailoverChainer is implemented
		here.
		
		Arguments
		
		  self -- this instance
		  
		  session -- session we belong to
		
		"""
		local = None
		hostlist = session.server_address
		if type(hostlist) != types.TupleType:
			return ConnectChainer.chainParent(self, session)

		if len(hostlist) == 1:
			session.server_address = hostlist[0]
			return ConnectChainer.chainParent(self, session)

		if session.service.snat:
			local = session.service.snat.performTranslation(session, local, NAT_SNAT)

		first_attempt = self.current_host % len(hostlist)
		stream = None
		remote = None
 		while stream == None and (remote == None or (self.current_host != first_attempt)):
			remote = hostlist[self.current_host]
			self.current_host = (self.current_host + 1) % len(hostlist)

			if session.service.dnat:
				remote = session.service.dnat.performTranslation(session, remote, NAT_DNAT)
			stream = self.establishConnection(session, local, remote)
		return stream

class SideStackChainer(AbstractChainer):
	"""Class encapsulating proxy chaining side-by-side.

	This class encapsulates a special case of chaining. Instead of 
	establishing a connection to some kind of server it creates
	another Proxy instance and connects the server side of the current
	and the client side of the new Proxy to each other.

	Attributes

	  right_class -- proxy class to side-stack

	  right_chainer -- chainer used by right_class to chain further

	"""
	def __init__(self, right_class, right_chainer = None):
		"""Constructor to initialize a SideStackChainer class.

		This constructor performs initialization of the
		SideStackChainer class.

		Arguments

		  right_class -- proxy class to side-stack

		  right_chainer -- chainer used by right_class to chain further
		"""
		AbstractChainer.__init__(self)
		self.right_class = right_class
		if right_chainer == None:
			right_chainer = ConnectChainer()
		self.right_chainer = right_chainer

	def chainParent(self, session):
		"""Overridden function to perform chaining.

		This function is called by a Proxy instance to establish its
		server side connection. Instead of connecting to a server
		this chainer creates another proxy instance and connects
		this new proxy with the current one.

		Arguments
		
		  self -- this instance
		  
		  session -- session we belong to
		
		"""
		try:
			fds = socketPair(AF_UNIX, SOCK_STREAM)
		except IOError:
			log(CORE_SESSION, 3, "%s: side stacking failed, socketPair failed;" % (session.session_id,))
			return None

		try:
			fds = [fds[0], fds[1]]
			session.server_fd = fds[0]
			session.server_stream = Stream(fds[0], session.owner.session_id + ": leftside")
			ss = StackedSession(session, self.right_chainer)
			fds[0] = -1
			ss.client_fd = fds[1]
			ss.client_stream = Stream(ss.client_fd, ss.session_id + ": rightside")
			fds[1] = -1
			log(CORE_SESSION, 4, "%s: Side-stacking proxy instance; server_fd='%d', client_fd='%d', proxy_class='%s'" % (session.session_id, session.server_fd, ss.client_fd, self.right_class.__name__))
			self.right_class(ss, 0)
		except:
			log(CORE_ERROR, 3, "%s: Side-stacking failed; proxy_class='%s'" % (session.session_id, self.right_class.__name__))
			if (fds[0] != -1):
				closeFd(fds[0])
			if (fds[1] != -1):
				closeFd(fds[1])
